/* ======================================================================== *\
!
! *
! * This file is part of MARS, the MAGIC Analysis and Reconstruction
! * Software. It is distributed to you in the hope that it can be a useful
! * and timesaving tool in analysing Data of imaging Cerenkov telescopes.
! * It is distributed WITHOUT ANY WARRANTY.
! *
! * Permission to use, copy, modify and distribute this software and its
! * documentation for any purpose is hereby granted without fee,
! * provided that the above copyright notice appear in all copies and
! * that both that copyright notice and this permission notice appear
! * in supporting documentation. It is provided "as is" without express
! * or implied warranty.
! *
!
!
!   Author(s): Thomas Bretz  12/2002 <mailto:tbretz@astro.uni-wuerzburg.de>
!
!   Copyright: MAGIC Software Development, 2000-2003
!
!
\* ======================================================================== */

/////////////////////////////////////////////////////////////////////////////
//
// MHCerPhotEvt
//
/////////////////////////////////////////////////////////////////////////////
#include "MHCerPhotEvt.h"

#include <TCanvas.h>

#include "MLog.h"
#include "MLogManip.h"

#include "MParList.h"
#include "MCerPhotEvt.h"
#include "MCamDisplay.h"

#include "MGeomCam.h"
#include "MGeomPix.h"

ClassImp(MHCerPhotEvt);

// --------------------------------------------------------------------------
//
// Reset all pixels to 0 and reset fEntries to 0.
//
void MHCerPhotEvt::Clear()
{
    fSum.Reset();
//    fSum.InitSize(577);
    for (int i=0; i<577; i++)
    {
        fSum.AddPixel(i, 0, 0);
        fSum[i].SetPixelUnused();
    }
    fSum.FixSize();
    fEntries = 0;
}

// --------------------------------------------------------------------------
//
// Initialize the name and title of the task.
// Resets the sum histogram
//
MHCerPhotEvt::MHCerPhotEvt(const char *name, const char *title)
    : fCam(NULL), fEvt(NULL), fDispl(NULL)
{
    //
    //   set the name and title of this object
    //
    fName  = name  ? name  : "MHCerPhotEvt";
    fTitle = title ? title : "Average of MCerPhotEvts";

    Clear();
}

// --------------------------------------------------------------------------
//
// Delete the corresponding camera display if available
//
MHCerPhotEvt::~MHCerPhotEvt()
{
    if (fDispl)
        delete fDispl;
}

// --------------------------------------------------------------------------
//
// Get the event (MCerPhotEvt) the histogram might be filled with. If
// it is not given, it is assumed, that it is filled with the argument
// of the Fill function.
// Looks for the camera geometry MGeomCam and resets the sum histogram.
//
Bool_t MHCerPhotEvt::SetupFill(const MParList *plist)
{
    fEvt = (MCerPhotEvt*)plist->FindObject("MCerPhotEvt");
    if (!fEvt)
        *fLog << warn << GetDescriptor() << ": No MCerPhotEvt available..." << endl;

    fCam = (MGeomCam*)plist->FindObject("MGeomCam");
    if (!fCam)
        *fLog << warn << GetDescriptor() << ": No MGeomCam found." << endl;

    Clear();

    return kTRUE;
}

// --------------------------------------------------------------------------
//
// Fill the histograms with data from a MCerPhotEvt-Container.
//
Bool_t MHCerPhotEvt::Fill(const MParContainer *par, const Stat_t w)
{
    const MCerPhotEvt *evt = par ? (MCerPhotEvt*)par : fEvt;
    if (!evt)
    {
        *fLog << err << dbginf << "No MCerPhotEvt found..." << endl;
        return kFALSE;
    }

    const UInt_t n = evt->GetNumPixels();

    for (UInt_t i=0; i<n; i++)
    {
        const MCerPhotPix &pix = (*evt)[i];

        const Int_t id = pix.GetPixId();

        fSum[id].SetPixelUsed();
        fSum[id].AddNumPhotons(pix.GetNumPhotons());
    }

    fEntries++;

    return kTRUE;
}

// --------------------------------------------------------------------------
//
// Scale the sum container with the number of entries
//
Bool_t MHCerPhotEvt::Finalize()
{
    fSum.Scale(fEntries);
    return kTRUE;
}

// --------------------------------------------------------------------------
//
// Draw the present 'fill status'
//
void MHCerPhotEvt::Draw(Option_t *)
{
    if (!fCam)
    {
        *fLog << warn << "WARNING - Cannot draw " << GetDescriptor() << ": No Camera Geometry available." << endl;
        return;
    }

    TVirtualPad *pad = gPad ? gPad : MakeDefCanvas(this, 750, 600);
    pad->SetBorderMode(0);

    AppendPad("");
}

// --------------------------------------------------------------------------
//
// If a camera display is not yet assigned, assign a new one.
//
void MHCerPhotEvt::Paint(Option_t *option)
{
    if (!fCam)
    {
        *fLog << warn << "WARNING - Cannot paint " << GetDescriptor() << ": No Camera Geometry available." << endl;
        return;
    }

    if (!fDispl)
        fDispl = new MCamDisplay(fCam);

    fDispl->FillPhotNum(fSum);
    fDispl->Paint();
}
